;________________________________________________________________________
;
;                                G P S . A S M
;________________________________________________________________________
;
; NMEA Receiving Routines
; by Colin Buckup N5GG n5gg@hotmail.com April 00
; based on the original code PICE90 for Position Sending with the PIC-E
; previously written by Byon Garrabrant  N6BG  byon@mail.com  Feb 99
; All copyrights that apply do Byon's code also apply to this version
;________________________________________________________________________
;
; This code take an incoming char stream of GPS NMEA data, and
; builds MIC-E compatable compressed data.  The incoming data is 
; buffered, so the data is valid at any time this routine is not running.
;________________________________________________________________________
;
; Revision
; v 1.0 - 04/06/00 - Colin
;
;________________________________________________________________________

#define		NMEA_STRING_LEN		D'22'
#define		MIC_E_DEST_LEN		D'7'
#define		MIC_E_INF_LEN		D'7'
#define		ASCII_OFFSET		D'28'
#define		VALID_FLAG			H'60'
#define		INVALID_FLAG		H'27'
#define		SPEED_TAB_MAX		D'6'
;________________________________________________________________________

	cblock				; variables
		NMEA_current_char	; the current incoming char
		NMEA_mode		; the current receive mode
		NMEA_index		; the current mode index
		NMEA_valid_timer	; the valid timer

		SPEED_TABLE		; the beacon rate table offset
		SPEED_NM		; the current speed in NM
		TAB_PTR		; auxiliary table pointer

		NMEA_STRING_BASE : NMEA_STRING_LEN ; NMEA message
		MIC_E_DEST_BASE : MIC_E_DEST_LEN   ; the outgoing dest call
		MIC_E_INF_BASE : MIC_E_INF_LEN	   ; the outgoing info
	endc
;________________________________________________________________________
; Various routines

NMEA_Get_100X				; multiplies the incoming w by 100
	andlw	H'3'			; assumes w < 3
	addwf	PCL, F
	dt	D'0', D'100',D'200'

NMEA_Get_10X				; multiplies the incoming w by 10
	andlw	H'F'			; assume w < 10
	addwf	PCL, F
	dt	D'0', D'10',D'20',D'30',D'40',D'50',D'60',D'70',D'80',D'90'

Get_NMEA_Title				; gets the NMEA title to watch for
	andlw	H'7'
	addwf	PCL, F
	dt	"$GPRMC,"

NMEA_Get_Mode_Size			; gets the max number of chars 
	addwf	PCL, F			; for each mode 0 .. 8
	dt	0,   0,   D'1', D'6', D'1', D'7', D'1', D'3', D'3'

NMEA_Get_Mode_Offset			; get the offset in NMEA_STRING
	addwf	PCL, F			; to store the current mode
	dt	0,   0,   D'0', D'1', D'7', D'8', D'15',D'16',D'19'

MIC_E_Default_Data			
	addwf	PCL, F			; to store the current mode
	dt	'E'<<1, 'R'<<1, 'R'<<1, 'O'<<1, 'R'<<1, '!'<<1, H'60' 
	dt	"NO GPS "

Get_Vmin			; gets the min speed in kn for a beacon rate table offset 
	addwf	PCL, F			; for each of the 7 table values 0 .. 6
	dt	D'0', D'2', D'5', D'12', D'22', D'39', D'68'

;________________________________________________________________________
; Routine to find out the right position update rate based on
; current conditions and stored peak speed during current interval

Find_update_rate
	movlw	SPEED_TAB_MAX		; load the offset for fastest update rate
	movwf	TAB_PTR			; store in temporary pointer
Range_loop
	movlw	HIGH Get_Vmin		; make sure that you'll point to right page...
	movwf	PCLATH
	movfw	TAB_PTR			; load intended offset
	call	Get_Vmin			; get the min speed for that update rate
	subwf SPEED_NM, W			; is current speed higher than min speed?
	bc	Right_range			; yes, then use this range
	decf	TAB_PTR, F			; no, so decrease range and try again
	bnz	Range_loop			; try again until look-up is finished
Right_range
	movfw	TAB_PTR			; load chosen pointer
	subwf	SPEED_TABLE, W		; is current speed higher that peak speed recorded?
	bc	Right_range_end		; no, so keep previous update rate
	movfw	TAB_PTR			; yes, then update to
	movwf	SPEED_TABLE			; new (faster) update rate
Right_range_end
	return
;________________________________________________________________________
;Convert ASCII to binary routines

NMEA_Read_2_Digit_Binary		; Reads a 2 digit ascii number
	movwf	FSR			; from RAM
	clrf	NMEA_mode		; start with 0
	goto	NMEA_Read_Binary_Continue
NMEA_Read_3_Digit_Binary		; Reads a 3 digit ascii number
	movwf	FSR			; from RAM
	movfw	INDF			; get hundreds digit
	call	NMEA_Get_100X		; multiply by 100
	movwf	NMEA_mode		; save it
	incf	FSR, F			; bump to next digit
NMEA_Read_Binary_Continue
	movfw	INDF			; get tens digit
	call	NMEA_Get_10X		; multiply by 10 
	addwf	NMEA_mode, F		; add to result
	incf	FSR, F			; bump to next digit
	movfw	INDF			; read ones digit
	andlw	H'0F'			; convert to binary
	addwf	NMEA_mode, W		; add to result
	return				; return in W
;________________________________________________________________________
; Initialization routines

NMEA_Init				
	clrf	NMEA_index
	clrf	SPEED_NM
NMEA_Init_Loop
	movfw	NMEA_index
	addlw	MIC_E_DEST_BASE
	movwf	FSR
	movfw	NMEA_index
	call	MIC_E_Default_Data
	movwf	INDF
	incf	NMEA_index, F
	movfw	NMEA_index
	xorlw	MIC_E_DEST_LEN + MIC_E_INF_LEN
	bnz	NMEA_Init_Loop		; if not complete, loop back
	goto	NMEA_Reset
;________________________________________________________________________
; Main NMEA routine

NMEA_Decode
	movwf	NMEA_current_char	; Save the current char
	tstf	NMEA_mode		; are we receiving data?
	bnz	NMEA_Receive		; yes, jump ahead

	movfw	NMEA_index		; compare incoming char
	call	Get_NMEA_Title		; to the string $GPRMC
	xorwf	NMEA_current_char, W	; if mismatch,
	bnz	NMEA_Reset		; reset, and continue looking	

	incf	NMEA_index, F		; else, bump to next char
	movlw	','			; if we didn't just get
	xorwf	NMEA_current_char, W	; the comma,
	skpz				;
	 return				; leave

	movlw	NMEA_STRING_BASE + D'16'; we received the complete title
	movwf	FSR			; flush RAM speed and bering
NMEA_Clear_Loop
	movlw	'0'			; by writing '0' into
	movwf	INDF			; those 6 locations
	incf	FSR, F
	movfw	FSR			; check for completion
	xorlw	NMEA_STRING_BASE + NMEA_STRING_LEN
	bnz	NMEA_Clear_Loop		; if not complete, loop back
	goto	NMEA_Next_Mode		; else goto next mode
;________________________________________________________________________
; Routine if valid data are being received

NMEA_Receive				; We are receiving the data
	movfw	NMEA_current_char	; Get the current char
	xorlw	','			; if comma
	bz	NMEA_Next_Mode		; goto next mode

	movfw	NMEA_current_char	; Get the current char
	xorlw	'*'			; compare to star
	skpnz				; if star
	 goto	NMEA_Reset		; reset receive

	movfw	NMEA_mode		; for the current mode
	call	NMEA_Get_Mode_Size	; compare the number of chars to get
	xorwf	NMEA_index, W		; with the index
	skpnz				; did we hit the end?
	 return				; leave (wait for next mode)

	movfw	NMEA_mode		; is the mode
	sublw	D'7'-1			; greater than 7
	bnc	NMEA_Get_Speed		; get speed/bering

NMEA_Get_Position			; get position and chars
	movfw	NMEA_current_char	; Get the current char
	xorlw	'.'			; compare to period
	skpnz				; if period,
	 return				; leave (wait for next char)
	
	movfw	NMEA_mode		; for the current mode
	call	NMEA_Get_Mode_Offset	; get to mode offset
	addwf	NMEA_index, W		; add the index
	addlw	NMEA_STRING_BASE	; received char
	movwf	FSR		
	movfw	NMEA_current_char	; and store the current char
	movwf	INDF
	incf	NMEA_index, F		; inc the index
	return
;________________________________________________________________________
; Get speed/bearing by shifting in the data

NMEA_Get_Speed				; Get Speed and Bering
	movfw	NMEA_current_char	; Get the current char
	xorlw	'.'			; compare to period
	skpnz				; if period,
	 goto	NMEA_Get_Speed_Complete	; end this get

	movfw	NMEA_mode		; for the current mode
	call	NMEA_Get_Mode_Offset	; get to mode offset
	addlw	NMEA_STRING_BASE	; add to base
	movwf	FSR			; save as fsr

	incf	FSR, F			
	movfw	INDF			; copy the tens digit
	decf	FSR, F			
	movwf	INDF			; to the hundreds digit
	incf	FSR, F
	incf	FSR, F
	movfw	INDF			; copy the ones digit
	decf	FSR, F
	movwf	INDF			; to the tens digit
	incf	FSR, F
	movfw	NMEA_current_char	; copy the current char
	movwf	INDF			; to the ones digit
	return				; done

NMEA_Get_Speed_Complete
	movfw	NMEA_mode		; for the current mode
	call	NMEA_Get_Mode_Size	; get the number of chars to get
	movwf	NMEA_index		; save as index so no more will 
	return				; be added and return
;________________________________________________________________________

NMEA_Next_Mode				; Goto the next NMEA mode
	clrf	NMEA_index		; reset the index
	incf	NMEA_mode, F		; inc the mode

	movfw	NMEA_mode		; check the mode
	xorlw	D'9'			; if the mode is not 9
	skpz				
	 return				; done

	call	MIC_E_Convert		; convert the data and reset

NMEA_Reset				; Reset back to watchin for title				
	clrf	NMEA_index		; clear index and
	clrf	NMEA_mode		; mode and 
	return				; return
;________________________________________________________________________
; Convert received data to MIC-E format

MIC_E_Convert				; Convert the received data to MIC-E
	movlw	NMEA_STRING_BASE	; set up initial FSR to begining
	movwf	FSR			; of the string

	movfw	INDF			; get the Valid char
	xorlw	'A'			; compare to 'A'
	movlw	INVALID_FLAG		; if not 'A'
	skpnz				; use INVALID_FLAG,
	 movlw	VALID_FLAG		; else use VALID_FLAG
	skpnz				; clear the valid timer if
	 clrf	NMEA_valid_timer	;  valid data
	movwf	MIC_E_INF_BASE + 0 	; store in INF_0
	incf	FSR, F			; goto next char in data

	clrf	NMEA_index		; reset index counter
MIC_E_Dest_Setup_Loop			; Copy the LAt to the Dest Callsign
	rlf	INDF, W			; read data, left shifted
	movwf	NMEA_mode		; save in temp storage

	movlw	MIC_E_DEST_BASE - NMEA_STRING_BASE - 1	
	addwf	FSR, F			; move to output

	movfw	NMEA_mode		; get value from temp storage
	andlw	B'00011110'		; mask off lower nibble
	iorlw	B'01100000'		; set bit 5 & 6

	movwf	INDF			; store result

	movlw	NMEA_STRING_BASE + 1 - MIC_E_DEST_BASE + 1
	addwf	FSR, F			; back down FSR and inc

	incf	NMEA_index, F		; update index count
	movfw	NMEA_index		; check it
	xorlw	6			; if not converted 6 yet
	bnz	MIC_E_Dest_Setup_Loop	; continue


	movlw	OPT_MESSAGE		; Copy the Message
	call	Get_Option
	movwf	NMEA_index		; bits to the high

	movlw	B'11000000'		; xor mask	

	btfss	NMEA_index, 0		; bits of the first
	 xorwf	MIC_E_DEST_BASE + 2, F	; three dest bytes
	btfss	NMEA_index, 1		; (inverted)
	 xorwf	MIC_E_DEST_BASE + 1, F
	btfss	NMEA_index, 2
	 xorwf	MIC_E_DEST_BASE + 0, F

	btfsc	NMEA_STRING_BASE + 7, 3	; update N/S 
	 xorwf	MIC_E_DEST_BASE + 3, F	; bit 3 is 1 for N, 0 for S

	btfsc	NMEA_STRING_BASE + 8, 0 ; update 100's long
	 xorwf	MIC_E_DEST_BASE + 4, F	; bit 0 is 1 for 1, 0 for 0

	btfsc	NMEA_STRING_BASE + D'15', 1 ; update E/W 
	 xorwf	MIC_E_DEST_BASE + 5, F	; bit 1 is 1 for W, 0 for E

	movlw	OPT_PATH		; get and save the path (SSID)
	call	Get_Option
	movwf	MIC_E_DEST_BASE + 6		
	clrc	
	rlf	MIC_E_DEST_BASE + 6, F 	; and shift left
	iorlw   H'E0'			; set Cmd/Response flag and both "rr" bits (srb)

	movlw	NMEA_STRING_BASE + D'9'	; get Long deg
	call	NMEA_Read_2_Digit_Binary
	addlw	ASCII_OFFSET		; add offset
	movwf	MIC_E_INF_BASE + 1	; and store

	movlw	NMEA_STRING_BASE + D'11'; get long min
	call	NMEA_Read_2_Digit_Binary
	addlw	ASCII_OFFSET		; add offset
	movwf	MIC_E_INF_BASE + 2	; and store

	movlw	NMEA_STRING_BASE + D'13'; get long hunds
	call	NMEA_Read_2_Digit_Binary
	addlw	ASCII_OFFSET		; add offset
	movwf	MIC_E_INF_BASE + 3	; and store

	movlw	NMEA_STRING_BASE + D'16'; get speed
	call	NMEA_Read_3_Digit_Binary	; as a 3 digit number

; Determine the correct update rate according to peak speed
	movwf	SPEED_NM			; store the current speed
	call	Find_update_rate		; get the right position update rate
	movfw	SPEED_NM			; load current speed and proceed
; resume conversion

	movwf	MIC_E_INF_BASE + 5	; store the miles
	clrf	MIC_E_INF_BASE + 4	; now compute miles div 10
	movlw	D'10'			; by subtracting 10 from the
MIC_E_Speed_Divide_Loop			; miles and counting the 
	incf	MIC_E_INF_BASE + 4, F	; subtractions
	subwf	MIC_E_INF_BASE + 5, F
	bc	MIC_E_Speed_Divide_Loop
	decf	MIC_E_INF_BASE + 4, F	; we will have subtracted 1 time			;(352)
	addwf	MIC_E_INF_BASE + 5, W	; too many, so fix up
	call	NMEA_Get_10X		; and save the remainder X 10
	movwf	MIC_E_INF_BASE + 5   	; back in INF_5

	movlw	ASCII_OFFSET + D'80'	; bump 4 by the ascii offset, 
	addwf	MIC_E_INF_BASE + 4, F	; plus add 800 to speed

	movlw	NMEA_STRING_BASE + D'19'; get bering div 100
	movwf	FSR			; by grabbing the 1 digit
	movfw	INDF
	andlw	H'0F'			; convert to binary
	addlw	ASCII_OFFSET + D'4'	; and add with the offset + 400
	addwf	MIC_E_INF_BASE + 5, F	; to INF_5

	movlw	NMEA_STRING_BASE + D'20'; get bering mod 100
	call	NMEA_Read_2_Digit_Binary; (the last two digits)
	addlw	ASCII_OFFSET		; add the offset
	movwf	MIC_E_INF_BASE + 6	; and store in INF_6

	
	movfw	NMEA_STRING_BASE + D'11'; if long min tens is 0
	andlw	H'0F'
	movlw	D'60'			; add 60
	skpnz
	 addwf	MIC_E_INF_BASE + 2, F

	movfw	NMEA_STRING_BASE + D'9'	; if long deg tens is not 0
	andlw	H'0F'
	skpz
	 return				; it's converted

	movlw	D'80'
 	addwf	MIC_E_INF_BASE + 1, F	; add 80
	btfsc	MIC_E_DEST_BASE + 4, 7	; if long deg hund
	 return				; it's converted
	
	movlw	D'10'
 	addwf	MIC_E_INF_BASE + 1, F	; add 10
	bsf	MIC_E_DEST_BASE + 4, 7	; set 100's
	bcf	MIC_E_DEST_BASE + 4, 6	; set 100's
	return				; it's converted
;________________________________________________________________________

